/******************************************************************************/
/* ST_ALLOC.C								      */
/******************************************************************************/

/* -----------------------------------------------------------------------------
 * General-purpose memory allocator, on the MWC arena model, with this added
 * feature:
 *
 * All blocks are coalesced when they're freed.  If this results in an arena
 * with only one block, and that free, it's returned to the OS.
 *
 * The functions here have the same names and bindings as the MWC memory
 * manager, which is the same as the UNIX names and bindings.
 *
 * DLF - 8/2/89	 Malloc bug with 1 byte mallocs fixed.
 */
/* -------------------------------------------------------------------------- */

#include "st_def.h"

/* -------------------------------------------------------------------------- */

/*
 * block header : every memory block has one.
 *
 * The size field is negative (i.e. -size) if the block is allocated.
 * Zero-length blocks are free; they hold space which might get coalesced
 * usefully later on.
 */

struct	block
	{
	struct	block	*next;	/* NULL for last guy */
	struct	block	*prev;	/* NULL for first guy */
	};

/*
 * arena header : every arena has one.
 *
 * Each arena is always completely filled with blocks; the first starts right
 * after this header.  There is also a block right at the end of the arena;
 * This is done since the blocks are doubly linked.
 */

struct	arena
	{
	struct	arena	*aptr;	/* actual pointer from Malloc() */
	size_t	size;		/* size of the arena */

	struct	block	*bptr;	/* First free block in arena */

	struct	arena	*next;	/* next arena in list */
	struct	arena	*prev;	/* previous arena in list */
	};

static struct arena *mem_a_first = NULL;

/*
 * "mem_cache_size" must be declared and initialized elsewhere
 */

static	size_t	mem_cache_size = 0x00010000L;

/*
 * If this pointer is set, that means a new document is being loaded and so
 * start with a new arena and always get memory from new arena in order to
 * avoid the scanning the old memory arenas.  This is done in order to
 * increase the efficiency of memory management unit.  Ofcourse there
 * is a tradeoff here.
 *
 * We use a flag here, which says whether we are following the normal memory
 * allocation routines (when FALSE) or the special memory allocation routines
 * which are used only while loading a document from disk (when TRUE).
 *
 */

static	int	mem_docu_arena_flag = FALSE;
static	struct arena	*mem_docu_arena_ptr = NULL;

/* ----------------------------------------------------------------------- */

static	struct	block	*alloc_arena( size_t );

/******************************************************************************/
/******************************************************************************/

/* -----------------------------------------------------------------------------
 */

static	struct	block	*alloc_arena(size)
register size_t	size;
{
register struct	arena	*a;
register struct	block	*b;
register struct	block	*c;
register size_t	temp;
register size_t	tsize;
struct	arena	*aa;

	if ((temp = mem_cache_size) < size)
	{
		temp = ((size / 0x00010000L) + 1) * 0x00010000L;
	}

/* Size of arena is mem needed plus size of arena structure and */
/* size of a block structure. If Malloc failed return failure */

	tsize = temp + sizeof(struct arena) + 2 * sizeof(struct block);

	a = (struct arena *)Malloc((long)tsize);
	if (a == NULL)
	{
		return(NULL);
	}

/* Set the pointer received as an even memory pointer (align to word boundary) */
/* Also save the original pointer received as 'aptr' in the arena structure */

	aa = (struct arena *)(((long)a + 1) & ~1L);
	aa->aptr = a;
	a = aa;

/* Establish the next and previous pointers in arena list */

	a->size = tsize;
	a->prev = NULL;
	if (a->next = mem_a_first)
	{
		mem_a_first->prev = a;
	}
	mem_a_first = a;

/* Establish a block in this arena, equal in size to available */
/* memory in this arena */

	b = (struct block *)(a+1);
	c = (struct block *)((long)(b+1) + temp);
	b->prev = NULL;
	b->next = c;
	c->next = NULL;
	c->prev = (struct block *)((long)b | 1L);

/* Set the first free block pointer in arena */

	a->bptr = b;
	return(b);
}

/******************************************************************************/
/******************************************************************************/

/* -----------------------------------------------------------------------------
 */

void	mem_set_load_alloc()
{
	if (mem_docu_arena_flag)
	{
		return;
	}
	free_arenas();
	mem_docu_arena_flag = TRUE;
	mem_docu_arena_ptr = NULL;
	return;
}

/* -----------------------------------------------------------------------------
 */

void	mem_reset_load_alloc()
{
	if (!mem_docu_arena_flag)
	{
		return;
	}

	free_arenas();
	mem_docu_arena_flag = FALSE;
	mem_docu_arena_ptr = NULL;
	return;
}

/******************************************************************************/
/******************************************************************************/

/* -----------------------------------------------------------------------------
 */

void	*malloc(size)
register size_t	size;
{
register struct	arena	*a;
register struct	block	*b;
register struct	block	*nb;
register size_t	temp;
register size_t	size_available;
register int	sizeblck;

/* Report failure if negative or zero size block is requested */

	if (size == 0L)
	{
		return(NULL);
	}

/* make sure size is even :  fail if it's negative zero */

	size = ((size + 1) & ~1L);
	sizeblck = sizeof(struct block);

/* If memory allocation requested from fresh arenas then get a new arena */
/* for the fresh memory allocation.  So if repeated request for memory */
/* allocation is received, then the request can be satisfied from this arena */
/* without scanning the arena for free blocks.  This way a lot of time is saved */
/* by getting the requested memory block in one step. */
/* */
/* In case, no more fresh arenas can be found, then get memory by scanning the */
/* existing arenas. In this case we can afford to loose efficiency, because the */
/* system is running short of memory */

	if (mem_docu_arena_flag)
	{

/* If next block in one such fresh arenas is available and its size is greater */
/* than the requested size, then go ahead and break the block tio satisfy the */
/* allocation request - it's a one step process */

		if ((a = mem_docu_arena_ptr) && (b = a->bptr))
		{
			size_available = (long)b->next - (long)(b+1);
			if (size_available >= size)
			{
				goto done;
			}
		}

/* If no big enough block found in the fresh arena, then allocate a new memory */
/* arena big enough to hold the requested memory size.  In case a fresh arena */
/* can be found, then start the time consuming process of searching free blocks */
/* in the existing arenas */

		b = alloc_arena(size);
		if (b)
		{
			mem_docu_arena_ptr = a = mem_a_first;

			size_available = (long)b->next - (long)(b+1);
			goto done;
		}
	}

/* Look into existing arenas for the free block */

	if (a = mem_a_first)	/* does any arena exist yet ? */
	{
		temp = size + sizeblck;
		while (a)
		{
			for (b = a->bptr ; b ; b = b->next)
			{
				if ((long)(b->prev) & 1L)
				{
					continue;
				}

				/* if free & big enough block, use it */

				size_available = (long)b->next - (long)b;
				if (size_available >= temp)
				{
					size_available -= (long)sizeblck;
					goto done;
				}
			}
			a = a->next;
		}
	}

/* If no block found, then allocate a new memory arena big enough to hold size */

	b = alloc_arena(size);
	if (b == NULL)
	{
		return(NULL);
	}

	a = mem_a_first;
	size_available = (long)b->next - (long)(b+1);

/* Split the block (if possible) so that it is an allocated block, size bytes */ 
/* long, and the rest of the original block becomes a new, free block, which is */
/* b->size - size - sizeof(struct block) bytes long */

done :

/* Set temp to size of remainder */

	if (size_available > (size + sizeblck))
	{

/* Block is big enough to cut : cut it */

		nb = (struct block *)((long)(b+1) + size);
		nb->next = b->next;
		nb->next->prev = (struct block *)((long)nb | ((long)(nb->next->prev) & 1L));
		b->next = nb;
		nb->prev = b;

/* If the first free block in arena was big enough to be broken into two blocks */
/* then set the next free block to 'nb' */

		if (a->bptr == b)
		{
			a->bptr = nb;
		}
	}
	else
	{

/* If the whole block is taken up and it was the first free block in arena, then */
/* look for next free block in this arena */

		if (a->bptr == b)
		{
			nb = b->next;
			while (nb && ((long)(nb->prev) & 1L))
			{
				nb = nb->next;
			}
			a->bptr = nb;
		}
	}

/* This means the block is allocated */

	b->prev = (struct block *)((long)b->prev | 1L);
	return((void *)(b+1));
}

/* -----------------------------------------------------------------------------
 */

void	*calloc(count, size)
size_t	count;
size_t	size;
{
register int	*ptr;
register void	*buf;
register size_t	lsize;

/* Allocate a block big enough to hold 'count' items each of 'size' bytes */

	lsize = count * size;
	buf = malloc(lsize);
	if (buf == NULL)
	{
		return(NULL);
	}

/* Clear the last odd byte if any */

	if (lsize & 1)
	{
		*((char *)buf + lsize - 1) = 0;
	}

/* Clear complete words & return */

	lsize >>= 1;
	ptr = (int *)buf;
	while (lsize --)
	{
		*ptr ++ = 0;
	}

	return((void *)buf);
}

/******************************************************************************/
/******************************************************************************/

/* -----------------------------------------------------------------------------
 */

void	free(fb)
void	*fb;
{
register struct	block	*b;
register struct	block	*nb;
register struct	arena	*a;

/* Set fb (and b) to header start; negate size */

	b = (struct block *)fb;
	-- b;

	if (!((long)(b->prev) & 1L))
	{

/* This isn't an allocated block! */

		return;
	}

/* Free the block */

	b->prev = (struct block *)((long)b->prev & (~1L));

/* Coalesce backwards if any prev and it's free */

	if ((nb = b->prev) && !((long)(nb->prev) & 1L))
	{
		b = b->next;
		nb->next = b;
		b->prev = (struct block *)((long)nb | ((long)(b->prev) & 1L));
		b = nb;
	}

/* Coalesce forwards if any next and it's free */

	if ((nb = b->next) && !((long)nb->prev & 1L))
	{
		nb = nb->next;
		b->next = nb;
		nb->prev = (struct block *)((long)b | ((long)(nb->prev) & 1L));
	}

/* Set the first free block pointer in the arena */
/* where the block just freed belongs */

	a = mem_a_first;
	while (a)
	{
		if (((long)b > (long)a->aptr) && ((long)b < ((long)a->aptr + a->size)))
		{
			break;
		}
		a = a->next;
	}

	if ((a->bptr > b) || (a->bptr == NULL))
	{
		a->bptr = b;
	}

	return;
}

/* -----------------------------------------------------------------------------
 */

void	free_arenas()
{
register struct	arena	*a, *na;
register struct	block	*b;

	a = mem_a_first;
	while (a)
	{
/* Get the pointer to first block in this arena 'a' */

		b = (struct block *)(a+1);

		if ((long)(b->prev) & 1L)
		{
			a = a->next;
			continue;
		}

		if (b->next->next)
		{
			a = a->next;
			continue;
		}

/* This arena now contains one free block : Mfree it! */

		if (na = a->next)
			na->prev = a->prev;

		if (a->prev)
			a->prev->next = na;
		else	mem_a_first = na;

		Mfree(a->aptr);
		a = na;
	}

	return;
}

/******************************************************************************/
/* End of ST_ALLOC.C							      */
/******************************************************************************/
